#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
__author__ = "Randall Balestriero"

import versioneer

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="symjax",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Randall Balestriero",
    author_email="randallbalestriero@gmail.com",
    description="A Symbolic JAX software",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SymJAX/SymJAX.git",
    packages=find_packages(exclude=["examples"]),
    classifiers=[
        "Natural Language :: English",
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
    ],
    python_requires=">=3.6",
    install_requires=["numpy", "scipy", "jax", "jaxlib", "networkx",],
    license="Apache-2.0",
)



import numpy

import symjax
from symjax import tensor
from ..base import gradients


class Optimizer:
    def __init__(self, *args, name=None, **kwargs):

        if name is None:
            name = self.__NAME__
        with symjax.Scope(name):
            self.create_updates(*args, **kwargs)

    def reset(self):
        if hasattr(self, "variables"):
            for var in self.variables:
                var.reset()

    @property
    def updates(self):
        if hasattr(self, "_update"):
            return self._updates
        else:
            self._updates = {}
            return self._updates

    def _get_grads(self, grads_or_loss, params):
        # get grads if given is loss
        if isinstance(grads_or_loss, tensor.Tensor):
            return gradients(grads_or_loss, params)
        else:
            return grads_or_loss

    def add_updates(self, update):
        self.updates.update(update)
        symjax.current_graph().add(update)


class SGD(Optimizer):
    """Gradient Descent Optimization

    Parameters
    ----------

    grads_or_loss: scalar tensor or list of gradients
        either the loss (scalar of Tensor type) to be differentied
        or the list of gradients already computed and possibly altered
        manually (such as clipping)

    params: list of parameters to update
        if grads_or_loss is al list then it should be ordered w.r.t. the
        given parameters

    learning_rate: constant or Tensor
        the learning rate use to update the parameters

    Attributes
    ----------

    updates: list of updates

    variables: list of variables

    """

    __NAME__ = "SGDOptimizer"

    def create_updates(self, grads_or_loss, learning_rate, params=None):

        if params is None:
            params = [v for k, v in get_graph().variables.items() if v.trainable]
        elif type(params) is not list and type(params) is not tuple:
            params = [params]

        grads = self._get_grads(grads_or_loss, params)

        if not numpy.isscalar(learning_rate) and not isinstance(
            learning_rate, tensor.Placeholder
        ):
            learning_rate = learning_rate()

        updates = dict()
        for param, grad in zip(params, grads):
            updates[param] = param - learning_rate * grad

        self.add_updates(updates)


class NesterovMomentum(Optimizer):
    """Nesterov momentum Optimization

    Parameters
    ----------

    grads_or_loss: scalar tensor or list of gradients
        either the loss (scalar of Tensor type) to be differentied
        or the list of gradients already computed and possibly altered
        manually (such as clipping)

    params: list of parameters to update
        if grads_or_loss is al list then it should be ordered w.r.t. the
        given parameters

    learning_rate: constant or Tensor
        the learning rate use to update the parameters

    Attributes
    ----------

    updates: list of updates

    variables: list of variables

    """

    __NAME__ = "NesterovMomentumOptimizer"

    def create_updates(self, grads_or_loss, learning_rate, momentum, params=None):

        if params is None:
            params = [v for k, v in get_graph().variables.items() if v.trainable]

        grads = self._get_grads(grads_or_loss, params)

        if not numpy.isscalar(learning_rate) and not isinstance(
            learning_rate, tensor.Placeholder
        ):
            learning_rate = learning_rate()

        updates = dict()
        variables = []
        for param, grad in zip(params, grads):
            velocity = tensor.Variable(
                numpy.zeros(param.shape, dtype=param.dtype), trainable=False
            )
            variables.append(velocity)
            update = param - learning_rate * grad
            x = momentum * velocity + update - param
            updates[velocity] = x
            updates[param] = momentum * x + update

        self.add_updates(updates)


class Adam(Optimizer):
    """Adaptive Gradient Based Optimization with renormalization

    Parameters
    ----------

    grads_or_loss: scalar tensor or list of gradients
        either the loss (scalar of Tensor type) to be differentied
        or the list of gradients already computed and possibly altered
        manually (such as clipping)

    params: list of parameters to update
        if grads_or_loss is al list then it should be ordered w.r.t. the
        given parameters

    learning_rate: constant or Tensor
        the learning rate use to update the parameters

    beta1: constant or Tensor
        the value of the exponential moving average of the average of the
        gradients through time (updates)

    beta2: constant or Tensor
        the value of the exponential moving average of the variance of the
        gradients through time

    Attributes
    ----------

    updates: list of updates

    variables: list of variables

    """

    __NAME__ = "AdamOptimizer"

    def create_updates(
        self,
        grads_or_loss,
        learning_rate,
        beta1=0.9,
        beta2=0.999,
        epsilon=1e-8,
        params=None,
    ):

        if params is None:
            params = symjax.get_variables(trainable=True)

        grads = self._get_grads(grads_or_loss, params)

        # get the learning rate
        if callable(learning_rate):
            learning_rate = learning_rate()

        updates = dict()
        for param, grad in zip(params, grads):
            m = symjax.nn.schedules.ExponentialMovingAverage(grad, beta1)[0]
            v = symjax.nn.schedules.ExponentialMovingAverage(grad ** 2, beta2)[0]
            update = m / (tensor.sqrt(v) + epsilon)
            updates[param] = param - learning_rate * update

        self.add_updates(updates)



import os
import pickle, gzip
import urllib.request
import numpy as np
import tarfile
import time
from tqdm import tqdm
from scipy.io.wavfile import read as wav_read


class speech_commands:
    """speech commands classification

    source: https://ai.googleblog.com/2017/08/launching-speech-commands-dataset.html

    The dataset has 65,000 one-second long utterances of 30 short
    words, by thousands of different people, contributed by
    members of the public through the AIY website. It’s released
    under a Creative Commons BY 4.0 license, and will continue to
    grow in future releases as more contributions are received.
    The dataset is designed to let you build basic but useful
    voice interfaces for applications, with common words like
    “Yes”, “No”, digits, and directions included. The
    infrastructure we used to create the data has been open
    sourced too, and we hope to see it used by the wider community
    to create their own versions, especially to cover underserved
    languages and applications.

    """

    name2class = {
        "blues": 0,
        "classical": 1,
        "country": 2,
        "disco": 3,
        "hiphop": 4,
        "jazz": 5,
        "metal": 6,
        "pop": 7,
        "reggae": 8,
        "rock": 9,
    }

    def download(path=None):
        if path is None:
            path = os.environ["DATASET_path"]
        path += "speech_commands/"
        t0 = time.time()

        print("Downloading speech commands")

        # Check if directory exists
        if not os.path.isdir(path):
            print("\tCreating Directory")
            os.mkdir(path)

        # Check if file exists
        if not os.path.exists(path + "speech_commands_v0.01.tar.gz"):
            url = "http://download.tensorflow.org/data/speech_commands_v0.01.tar.gz"
            urllib.request.urlretrieve(url, path + "speech_commands_v0.01.tar.gz")

    def load(path=None):

        if path is None:
            path = os.environ["DATASET_PATH"]
        speech_commands.download(path)

        t0 = time.time()

        print("Loading speech command")

        tar = tarfile.open(
            path + "speech_commands/speech_commands_v0.01.tar.gz", "r:gz"
        )

        # Load train set
        wavs = list()
        labels = list()
        noises = list()
        noise_labels = list()
        names = tar.getmembers()
        for name in tqdm(names, ascii=True):
            if "wav" not in name.name:
                continue
            f = tar.extractfile(name.name)  # .read()
            wav = wav_read(f)[1]
            if "noise" in name.name:
                noises.append(wav)
                noise_labels.append(name.name.split("/")[-1])
            else:
                left = 16000 - len(wav)
                to_pad = left // 2
                wavs.append(np.pad(wav, [[to_pad, left - to_pad]]))
                labels.append(name.name.split("/")[-2])
        labels = np.array(labels)
        unique_labels = np.unique(labels)
        y = np.squeeze(
            np.array(
                [np.nonzero(label == unique_labels)[0] for label in labels]
            ).astype("int32")
        )

        data = {
            "wavs": np.array(wavs).astype("float32"),
            "labels": y,
            "names": labels,
            "noises": noises,
            "noises_labels": noises_labels,
            "INFOS": speech_commands.__doc__,
        }

        print("Dataset speech commands loaded in{0:.2f}s.".format(time.time() - t0))

        return data



import io
import os
import time
import urllib.request
import zipfile

import numpy as np
from scipy.io.wavfile import read as wav_read
from tqdm import tqdm


class audiomnist:
    """digit recognition
        https://github.com/soerenab/AudioMNIST

    A simple audio/speech dataset consisting of recordings of spoken digits in
    wav files at 8kHz. The recordings are trimmed so that they have near
    minimal silence at the beginnings and ends.

    FSDD is an open dataset, which means it will grow over time as data is
    contributed. In order to enable reproducibility and accurate citation the
    dataset is versioned using Zenodo DOI as well as git tags.

    Current status

        4 speakers
        2,000 recordings (50 of each digit per speaker)
        English pronunciations
    """

    def download(path):

        # Check if directory exists
        if not os.path.isdir(path + "audiomnist"):
            print("Creating audiomnist Directory")
            os.mkdir(path + "audiomnist")
        url = "https://github.com/soerenab/AudioMNIST/archive/master.zip"
        # Check if file exists
        if not os.path.exists(path + "audiomnist/data.zip"):
            td = time.time()
            print("Downloading Audio-MNIST")
            urllib.request.urlretrieve(url, path + "audiomnist/data.zip")
            print("Done in {} s.".format(time.time() - td))

    def load(path=None, subsample=1):

        if path is None:
            path = os.environ["DATASET_PATH"]
        audiomnist.download(path)
        t0 = time.time()

        # load wavs
        f = zipfile.ZipFile(path + "audiomnist/data.zip")
        wavs = list()
        digits = list()
        speakers = list()
        N = 0
        for filename in tqdm(f.namelist(), ascii=True):
            if ".wav" not in filename:
                continue
            filename_end = filename.split("/")[-1]
            digits.append(int(filename_end.split("_")[0]))
            speakers.append(int(filename_end.split("_")[1]) - 1)
            wavfile = f.read(filename)
            byt = io.BytesIO(wavfile)
            wavs.append(wav_read(byt)[1].astype("float32")[::subsample])
            N = max(N, len(wavs[-1]))

        digits = np.array(digits)
        speakers = np.array(speakers)
        all_wavs = np.zeros((len(wavs), N))
        for i in range(len(wavs)):
            left = (N - len(wavs[i])) // 2
            all_wavs[i, left : left + len(wavs[i])] = wavs[i]
        print("Audio-MNIST loaded in {} s.".format(time.time() - t0))
        return all_wavs, digits, speakers



import os
from zipfile import ZipFile
import urllib.request
import numpy as np
import scipy.misc

# from PIL import Image


class tinyimagenet:
    """
    Tiny Imagenet has 200 classes. Each class has 500 training images, 50
    validation images, and 50 test images. We have released the training and
    validation sets with images and annotations. We provide both class labels an
    bounding boxes as annotations; however, you are asked only to predict the
    class label of each image without localizing the objects. The test set is
    released without labels. You can download the whole tiny ImageNet dataset
    here.
    """

    def download(path):

        path = os.environ["DATASET_path"]

        if not os.path.isdir(path + "tinyimagenet"):
            os.mkdir(path + "tinyimagenet")

        if not os.path.exists(path + "tinyimagenet/tiny-imagenet-200.zip"):
            url = "http://cs231n.stanford.edu/tiny-imagenet-200.zip"
            urllib.request.urlretrieve(url, path + "tinyimagenet/tiny-imagenet-200.zip")

        return dataset

    def load_tinyimagenet(path=None):

        if path is None:
            path = os.environ["DATASET_path"]

        download(path)

        # Loading the file
        f = ZipFile(path + "tinyimagenet/tiny-imagenet-200.zip", "r")
        names = [name for name in f.namelist() if name.endswith("JPEG")]
        val_classes = np.loadtxt(
            f.open("tiny-imagenet-200/val/val_annotations.txt"),
            dtype=str,
            delimiter="\t",
        )
        val_classes = dict(
            [(a, b) for a, b in zip(val_classes[:, 0], val_classes[:, 1])]
        )
        x_train, x_test, x_valid, y_train, y_test, y_valid = [], [], [], [], [], []
        for name in names:
            if "train" in name:
                classe = name.split("/")[-1].split("_")[0]
                x_train.append(
                    scipy.misc.imread(
                        f.open(name), flatten=False, mode="RGB"
                    ).transpose((2, 0, 1))
                )
                y_train.append(classe)
            if "val" in name:
                x_valid.append(
                    scipy.misc.imread(
                        f.open(name), flatten=False, mode="RGB"
                    ).transpose((2, 0, 1))
                )
                arg = name.split("/")[-1]
                print(val_classes[arg])
                y_valid.append(val_classes[arg])
            if "test" in name:
                x_test.append(
                    scipy.misc.imread(
                        f.open(name), flatten=False, mode="RGB"
                    ).transpose((2, 0, 1))
                )

        return x_train, y_train, x_valid, y_valid, x_test



from symjax import tensor as T
from . import ops_nn as nn
from symjax.nn import initializers, schedules
import symjax
import numpy
import jax


def _is_shape(x):
    """utility function that checks if the input is a shape or not"""
    if not hasattr(x, "__len__"):
        return False
    if isinstance(x[0], int):
        return True
    elif isinstance(x[0], float):
        raise RuntimeError("invalid float value in shape")


def forward(input, layers):
    """ perform a forward path in the layers given an input

    utility function

    Parameters
    ----------

    input: Tensor
        the tensor to be forwarded across the given layers

    layers: list of layers
        the layers to forward the input in a consecutive manner
    """
    outputs = [layers[0].forward(input)]
    for layer in layers[1:]:
        outputs.append(layer.forward(outputs[-1]))
    return outputs


class Layer(T.Op):
    def __init__(self, *args, name=None, **kwargs):

        if name is None:
            name = self.__NAME__

        with symjax.Scope(name):
            output = self.forward(*args, **kwargs)
            super().__init__(output, 0, _jax_function=jax.numpy.add)

    def variables(self, trainable=True):
        if not hasattr(self, "_variables"):
            return []
        if trainable is not None:
            return [v for v in self._variables if v.trainable == trainable]
        else:
            return self._variables

    def init_input(self, input_or_shape):
        if _is_shape(input_or_shape):
            self.input = T.Placeholder(input_or_shape, "float32")
        else:
            self.input = input_or_shape

    def create_tensor(self, tensor_or_func, shape, dtype=None):
        if not callable(tensor_or_func):
            if tensor_or_func is None:
                return None
            assert tensor_or_func.shape == shape
            if tensor_or_func.dtype != dtype and dtype is not None:
                return tensor_or_func.astype(dtype)
            else:
                return tensor_or_func
        if dtype is None:
            dtype = "float32"
        try:
            tensor = tensor_or_func(shape=shape, dtype=dtype)
        except TypeError:
            tensor = tensor_or_func(shape=shape).astype(dtype)
        return tensor

    def create_variable(self, name, tensor_or_func, shape, trainable, dtype=None):
        if tensor_or_func is None:
            self.__dict__[name] = None
            return
        t = self.create_tensor(tensor_or_func, shape, dtype)

        if not trainable:
            self.__dict__[name] = t
        else:
            self.__dict__[name] = T.Variable(t, name=name, trainable=True)
            self.add_variable(self.__dict__[name])

    @property
    def updates(self):
        if not hasattr(self, "_updates"):
            self._updates = {}
        return self._updates

    def add_update(self, update):
        symjax.current_graph().add(update)

    def add_variable(self, variable):
        if not hasattr(self, "_variables"):
            self._variables = list()
        self._variables.append(variable)

    def forward(self):
        pass


class Identity(Layer):

    __NAME__ = "Identity"

    def forward(self, input):
        return input


class Upsample1D(Layer):

    __NAME__ = "Upsample1D"

    def forward(self, input, repeat, axis=-1, mode="constant", value=0.0):
        return T.interpolation.upsample_1d(
            input, repeat=repeat, axis=axis, mode=mode, value=value,
        )


class Upsample2D(Layer):

    __NAME__ = "Upsample2D"

    def forward(self, input, repeat, axis, mode="constant", value=0.0):
        p1 = T.upsample_1d(
            input, repeat=repeat[0], axis=axis[0], mode=mode, value=value,
        )
        p2 = T.upsample_1d(p1, repeat=repeat[1], axis=axis[1], mode=mode, value=value,)
        return p2


class Dense(Layer):
    """Fully-connected/Dense layer

    perform a dense matrix multiplication and bias shifting of the
    input
    """

    __NAME__ = "Dense"

    def forward(
        self,
        input,
        units,
        W=initializers.he,
        b=numpy.zeros,
        trainable_W=True,
        trainable_b=True,
    ):

        self.create_variable(
            "W", W, (numpy.prod(input.shape[1:]), units), trainable=trainable_W,
        )
        self.create_variable("b", b, (units,), trainable=trainable_b)

        if numpy.prod(input.shape[1:]) != self.W.shape[0]:
            raise RuntimeError("input to Dense layer {} has different dim".format(self))

        if self.b is not None:
            return T.dot(T.flatten2d(input), self.W) + self.b
        else:
            return T.dot(T.flatten2d(input), self.W)


class Conv1D(Layer):
    """1-D (time) convolution

    """

    __NAME__ = "Conv1D"

    def forward(
        self,
        input,
        n_filters,
        filter_length,
        W=initializers.he,
        b=numpy.zeros,
        stride=1,
        padding="VALID",
        trainable_W=True,
        trainable_b=True,
        input_dilations=None,
        filter_dilations=None,
    ):

        if numpy.isscalar(input_dilations):
            input_dilations = (input_dilations,) * 2
        self.input_dilation = input_dilations
        self.filter_dilation = filter_dilations
        self.stride = stride
        self.padding = padding

        self.create_variable(
            "W", W, (n_filters, input.shape[1], filter_length), trainable=trainable_W,
        )
        self.create_variable("b", b, (n_filters,), trainable=trainable_b)
        conv = nn.convNd(
            input,
            self.W,
            strides=self.stride,
            padding=self.padding,
            input_dilation=self.input_dilation,
            filter_dilation=self.filter_dilation,
        )
        if hasattr(self, "b"):
            return conv + self.b[:, None]
        else:
            return conv


class Conv2DTranspose(Layer):
    """2-D (spatial) convolution

    """

    __NAME__ = "Conv2DTranspose"

    def forward(
        self,
        input_or_shape,
        n_filters,
        filter_shape,
        pad="VALID",
        strides=1,
        W=initializers.he,
        b=numpy.zeros,
        trainable_W=True,
        trainable_b=True,
        transpose_W=True,
        filter_dilations=None,
    ):

        self.init_input(input_or_shape)
        self.transpose_W = transpose_W
        self.filter_dilation = filter_dilations
        self.strides = strides
        self.pad = pad

        self.create_variable(
            "W",
            W,
            (input.shape[1], n_filters) + tuple(filter_shape),
            trainable=trainable_W,
        )
        self.create_variable("b", b, (n_filters,), trainable=trainable_b)

        conv = T.convNd_transpose(
            input,
            self.W,
            strides=self.strides,
            padding=self.pad,
            transpose_kernel=self.transpose_W,
            filter_dilation=self.filter_dilation,
        )

        return conv + self.b.reshape((-1, 1, 1))


class Conv2D(Layer):
    """2-D (spatial) convolution

    """

    __NAME__ = "Conv2D"

    def forward(
        self,
        input,
        n_filters,
        filter_shape,
        pad="VALID",
        strides=1,
        W=initializers.he,
        b=numpy.zeros,
        trainable_W=True,
        trainable_b=True,
        input_dilations=None,
        filter_dilations=None,
    ):

        self.input_dilation = input_dilations
        self.filter_dilation = filter_dilations
        self.strides = strides
        self.pad = pad

        self.create_variable(
            "W",
            W,
            (n_filters, input.shape[1]) + tuple(filter_shape),
            trainable=trainable_W,
        )
        self.create_variable("b", b, (n_filters,), trainable=trainable_b)

        conv = nn.convNd(
            input,
            self.W,
            strides=self.strides,
            padding=self.pad,
            input_dilation=self.input_dilation,
            filter_dilation=self.filter_dilation,
        )
        if self.b is not None:
            return conv + self.b.reshape((-1, 1, 1))
        else:
            return conv


class Pool1D(Layer):
    """2-D (spatial) pooling

    """

    __NAME__ = "Pool1D"

    def forward(self, input_or_shape, pool_shape, pool_type="MAX", strides=None):

        self.init_input(input_or_shape)
        self.pool_type = pool_type
        self.pool_shape = (1, 1, pool_shape)
        if strides is None:
            self.strides = self.pool_shape
        else:
            self.strides = (1, 1, strides)

        return T.poolNd(
            input, self.pool_shape, strides=self.strides, reducer=self.pool_type,
        )


class Pool2D(Layer):
    """2-D (spatial) pooling

    """

    __NAME__ = "Pool2D"

    def forward(self, input, pool_shape, pool_type="MAX", strides=None):

        self.pool_type = pool_type
        self.pool_shape = (1, 1, pool_shape[0], pool_shape[1])
        if strides is None:
            self.strides = self.pool_shape
        else:
            if hasattr(strides, "__len__"):
                self.strides = (1, 1, strides[0], strides[1])
            else:
                self.strides = (1, 1, strides, strides)

        return nn.poolNd(
            input, self.pool_shape, strides=self.strides, reducer=self.pool_type,
        )


class Dropout(Layer):

    """binary mask onto the input

    Parameters
    ----------

    input_or_shape: shape or Tensor
        the layer input or shape

    p: float (0<=p<=1)
        the probability to drop the value

    deterministic: bool or Tensor
        the state of the layer

    seed: seed
        the RNG seed

    Returns
    -------

    output: the layer output

    """

    __NAME__ = "Dropout"

    def forward(self, input, p, deterministic, seed=None):

        self.p = p
        self.mask = T.random.bernoulli(shape=input.shape, p=p, seed=seed)

        return T.where(deterministic, input, self.mask * input)


class RandomFlip(Layer):

    """
    random axis flip on the input

    Random layer that will randomly flip the axis of the input.
    Note that all the involved
    operations are GPU compatible and allow for backpropagation

    Parameters
    ----------

    input_or_shape: shape or Tensor
        the input of the layer or the shape of the layer input

    crop_shape: shape
        the shape of the cropped part of the input. It must have the same
        length as the input shape  minus one for the first dimension

    deterministic: bool or Tensor
        if the layer is in deterministic mode or not

    padding: shape
        the amount of padding to apply on each dimension (except the first
        one) each dimension should have a couple for the before and
        after padding parts

    seed: seed (optional)
        to control reproducibility

    Returns
    -------

    output: the output tensor which containts the internal variables
    """

    __NAME__ = "RandomFlip"

    def forward(self, input, p, axis, deterministic, seed=None):

        self.p = p
        self.axis = axis
        extra_dims = input.ndim - 1
        self.flip = T.random.bernoulli(
            shape=(input.shape[0],) + (1,) * extra_dims, p=p, seed=seed
        )

        dirac = T.cast(deterministic, "float32")

        flipped_input = T.where(self.flip, T.flip(input, axis), input)

        return input * dirac + flipped_input * (1 - dirac)


class RandomCrop(Layer):

    """
    random crop selection form the input

    Random layer that will select a window of the input based on the given
    parameters, with the possibility to first apply a padding. This layer is
    commonly used as a data augmentation technique and positioned at the
    beginning of the deep network topology. Note that all the involved
    operations are GPU compatible and allow for backpropagation

    Parameters
    ----------

    input_or_shape: shape or Tensor
        the input of the layer or the shape of the layer input

    crop_shape: shape
        the shape of the cropped part of the input. It must have the same
        length as the input shape  minus one for the first dimension

    deterministic: bool or Tensor
        if the layer is in deterministic mode or not

    padding: shape
        the amount of padding to apply on each dimension (except the first
        one) each dimension should have a couple for the before and
        after padding parts

    seed: seed (optional)
        to control reproducibility

    Returns
    -------

    output: the output tensor which containts the internal variables

    """

    __NAME__ = "RandomCrop"

    def forward(self, input, crop_shape, deterministic, padding=0, seed=None):

        self.crop_shape = crop_shape
        # if given only a scalar
        if not hasattr(padding, "__len__"):
            self.pad_shape = [(padding, padding)] * (input.shape - 1)
        # else
        else:
            self.pad_shape = [
                (pad, pad) if not hasattr(pad, "__len__") else pad for pad in padding
            ]

        assert len(self.pad_shape) == len(self.crop_shape)
        assert len(self.pad_shape) == (len(input.shape) - 1)

        self.start_indices = list()
        self.fixed_indices = list()
        for i, (pad, dim, crop) in enumerate(
            zip(self.pad_shape, input.shape[1:], self.crop_shape)
        ):
            maxval = pad[0] + pad[1] + dim - crop
            assert maxval >= 0
            self.start_indices.append(
                T.random.randint(
                    minval=0,
                    maxval=maxval,
                    shape=(input.shape[0], 1),
                    dtype="int32",
                    seed=seed + i if seed is not None else seed,
                )
            )

            self.fixed_indices.append(
                T.ones((input.shape[0], 1), "int32") * (maxval // 2)
            )
        self.start_indices = T.concatenate(self.start_indices, 1)
        self.fixed_indices = T.concatenate(self.fixed_indices, 1)

        dirac = T.cast(deterministic, "float32")

        # pad the input
        pinput = T.pad(input, [(0, 0)] + self.pad_shape)

        routput = T.stack(
            [
                T.dynamic_slice(pinput[n], self.start_indices[n], self.crop_shape)
                for n in range(input.shape[0])
            ],
            0,
        )
        doutput = T.stack(
            [
                T.dynamic_slice(pinput[n], self.fixed_indices[n], self.crop_shape)
                for n in range(input.shape[0])
            ],
            0,
        )

        return doutput * dirac + (1 - dirac) * routput


class BatchNormalization(Layer):
    """
    batch-normalization layer


    Parameters:
    -----------

    input_or_shape: shape or Tensor
        the layer input tensor or shape

    axis: list or tuple of ints
        the axis to normalize on. If using BN on a dense layer then
        axis should be [0] to normalize over the samples. If the layer
        if a convolutional layer with data format NCHW then axis should
        be [0, 2, 3] to normalize over the samples and spatial dimensions
        (commonly done)

    deterministic: bool or Tensor
        controlling the state of the layer

    const: float32 (optional)
        the constant used in the standard deviation renormalization

    beta1: flaot32 (optional)
        the parameter for the exponential moving average of the mean

    beta2: float32 (optional)
        the parameters for the exponential moving average of the std

    Returns
    -------

    output: the layer output with attributes given by the layer options

    """

    __NAME__ = "BatchNormalization"

    def forward(
        self,
        input,
        axis,
        deterministic,
        const=1e-4,
        beta1=0.9,
        beta2=0.9,
        W=T.ones,
        b=T.zeros,
        trainable_W=True,
        trainable_b=True,
    ):

        self.beta1 = beta1
        self.beta2 = beta2
        self.const = const
        self.axis = axis
        self.deterministic = deterministic

        parameter_shape = [
            input.shape[i] if i in axis else 1 for i in range(input.ndim)
        ]
        reduce_axes = [i for i in range(input.ndim) if i not in axis]

        self.create_variable("W", W, parameter_shape, trainable=trainable_W)
        self.create_variable("b", b, parameter_shape, trainable=trainable_b)

        input_mean = T.mean(input, reduce_axes, keepdims=True)
        input_inv_std = 1 / (T.std(input, reduce_axes, keepdims=True) + const)

        self.avg_mean = schedules.ExponentialMovingAverage(input_mean, beta1)[1]
        self.avg_inv_std = schedules.ExponentialMovingAverage(input_inv_std, beta2)[1]

        use_mean = T.where(deterministic, self.avg_mean, input_mean)
        use_inv_std = T.where(deterministic, self.avg_inv_std, input_inv_std)
        W = self.W or 1.0
        b = self.b if self.b is not None else 0.0
        return W * (input - use_mean) * use_inv_std + b


class RNN(Layer):

    __NAME__ = "BatchNormalization"

    @staticmethod
    def gate(h, x, W, H, b, sigma):
        ht = sigma(T.dot(x, W) + b + T.dot(h, H))
        return ht, ht

    def forward(
        self,
        sequence,
        init_h,
        units,
        W=initializers.he,
        H=initializers.he,
        b=T.zeros,
        trainable_W=True,
        trainable_H=True,
        trainable_b=True,
        activation=nn.sigmoid,
        only_last=False,
    ):

        self.create_variable("W", W, (sequence.shape[2], units), trainable=trainable_W)
        self.create_variable("H", H, (units, units), trainable=trainable_H)
        self.create_variable("b", b, (units), trainable=trainable_b)

        last, output = T.scan(
            lambda h, x, W, H, b: self.gate(h, x, W, H, b, activation),
            init=init_h,
            sequences=[sequence.transpose((1, 0, 2))],
            non_sequences=[self.W, self.H, self.b],
        )
        if only_last:
            return last
        else:
            return output.transpose((1, 0, 2))


class GRU(Layer):

    __NAME__ = "BatchNormalization"

    @staticmethod
    def full_gate(h, x, Wh, Uh, bh, Wz, Uz, bz, Wr, Ur, br, sigma, phi):
        zt = sigma(T.dot(x, Wz) + bz + T.dot(h, Uz))
        rt = sigma(T.dot(x, Wr) + br + T.dot(h, Ur))
        h_hat = phi(T.dot(x, Wh) + bh + T.dot(h * rt, Uh))
        ht = (1 - zt) * h + zt * h_hat
        return ht, ht

    @staticmethod
    def minimal_gate(h, x, Wh, Uh, bh, Wz, Uz, bz, sigma, phi):
        ft = sigma(T.dot(x, Wz) + bz + T.dot(h, Uz))
        h_hat = phi(T.dot(x, Wh) + bh + T.dot(h * ft, Uh))
        ht = (1 - ft) * h + ft * h_hat
        return ht, ht

    def forward(
        self,
        sequence,
        init_h,
        units,
        Wh=initializers.he,
        Uh=initializers.he,
        bh=T.zeros,
        Wz=initializers.he,
        Uz=initializers.he,
        bz=T.zeros,
        Wr=initializers.he,
        Ur=initializers.he,
        br=T.zeros,
        trainable_Wh=True,
        trainable_Uh=True,
        trainable_bh=True,
        trainable_Wz=True,
        trainable_Uz=True,
        trainable_bz=True,
        trainable_Wr=True,
        trainable_Ur=True,
        trainable_br=True,
        activation=nn.sigmoid,
        phi=T.tanh,
        only_last=False,
        gate="minimal",
    ):

        self.create_variable(
            "Wh", Wh, (sequence.shape[2], units), trainable=trainable_Wh
        )
        self.create_variable("Uh", Uh, (units, units), trainable=trainable_Uh)
        self.create_variable("bh", bh, (units), trainable=trainable_bh)

        self.create_variable(
            "Wz", Wz, (sequence.shape[2], units), trainable=trainable_Wz
        )
        self.create_variable("Uz", Uz, (units, units), trainable=trainable_Uz)
        self.create_variable("bz", bz, (units), trainable=trainable_bz)

        if gate == "full":
            self.create_variable(
                "Wr", Wr, (sequence.shape[2], units), trainable=trainable_Wr
            )
            self.create_variable("Ur", Ur, (units, units), trainable=trainable_Ur)
            self.create_variable("br", br, (units), trainable=trainable_br)

        if gate == "minimal":

            def fn(h, x, Wh, Uh, bh, Wz, Uz, bz):
                return self.minimal_gate(h, x, Wh, Uh, bh, Wz, Uz, bz, activation, phi)

        elif gate == "full":

            def fn(h, x, Wh, Uh, bh, Wz, Uz, bz):
                return self.full_gate(
                    h, x, Wh, Uh, bh, Wz, Uz, bz, Wr, Ur, br, activation, phi
                )

        last, output = T.scan(
            fn,
            init=init_h,
            sequences=[sequence.transpose((1, 0, 2))],
            non_sequences=[self.W, self.H, self.b],
        )
        if only_last:
            return last
        else:
            return output.transpose((1, 0, 2))



import os
import gzip
import urllib.request
import numpy as np
import time
import zipfile
import io
from scipy.io.wavfile import read as wav_read
from tqdm import tqdm


class vocalset:
    """singer/technique/vowel of singing voices
    
    source: https://zenodo.org/record/1442513#.W7OaFBNKjx4

    We present VocalSet, a singing voice dataset consisting of 10.1 hours
    of monophonic recorded audio of professional singers demonstrating both
    standard and extended vocal techniques on all 5 vowels. Existing
    singing voice datasets aim to capture a focused subset of singing
    voice characteristics, and generally consist of just a few singers.
    VocalSet contains recordings from 20 different singers (9 male, 11
    female) and a range of voice types.  VocalSet aims to improve the
    state of existing singing voice datasets and singing voice research by
    capturing not only a range of vowels, but also a diverse set of voices
    on many different vocal techniques, sung in contexts of scales,
    arpeggios, long tones, and excerpts.
    """

    def download(path):

        if path is None:
            path = os.environ["DATASET_path"]

        # Load the dataset (download if necessary) and set
        # the class attributes.

        t = time.time()

        if not os.path.isdir(path + "vocalset"):

            print("\tCreating Directory")
            os.mkdir(path + "vocalset")

        if not os.path.exists(path + "vocalset/VocalSet11.zip"):
            print("Downloading Vocalset")
            url = "https://zenodo.org/record/1442513/files/VocalSet11.zip?download=1"
            urllib.request.urlretrieve(url, path + "vocalset/VocalSet11.zip")

            print("vocalset downloaded in {} sec.".format(time.time() - t))

    def load(path=None):
        """
        Parameters
        ----------

        path: str (optional)
            a string where to load the data and download if not present

        Returns
        -------

        singers: list
            the list of singers as strings, 11 males and 9 females as in male1,
            male2, ...

        genders: list
            the list of genders of the singers as in male, male, female, ...

        vowels: list
            the vowels being pronunced

        data: list
            the list of waveforms, not all equal length

        """
        if path is None:
            path = os.environ["DATASET_PATH"]

        vocalset.download(path)
        t = time.time()

        # load wavs
        f = zipfile.ZipFile(path + "vocalset/VocalSet11.zip")

        # init. the data array
        singers = []
        genders = []
        vowels = []
        #        techniques = []
        data = []
        for filename in tqdm(f.namelist(), ascii=True):
            if ".wav" not in filename or "excerpts" in filename or "_" == filename[0]:
                continue
            vowel = filename[-5]
            if vowel not in ["a", "e", "i", "o", "u"]:
                continue
            vowels.append(vowel)
            bytes_ = io.BytesIO(f.read(filename))
            data.append(wav_read(bytes_)[1].astype("float32"))
            split = filename.split("/")
            genders.append("".join(x for x in split[1] if x.isalpha()))
            singers.append(split[1])
        #            techniques.append(split[-1][3:-6])

        return singers, genders, vowels, data



import numpy as np


# https://github.com/deepmind/dsprites-dataset
class dsprites:
    """greyscale image classification and disentanglement

    This dataset consists of 737,280 images of 2D shapes, procedurally generated
    from 5 ground truth independent latent factors, controlling the shape, scale,
    rotation and position of a sprite. This data can be used to assess the
    disentanglement properties of unsupervised learning methods.

    dSprites is a dataset of 2D shapes procedurally generated from 6 ground
    truth independent latent factors. These factors are color, shape, scale,
    rotation, x and y positions of a sprite.

    All possible combinations of these latents are present exactly once,
    generating N = 737280 total images.
    """

    def download(path):

        """
        Download the MNIST dataset and store the result into the given
        path

        Parameters
        ----------

            path: str
                the path where the downloaded files will be stored. If the
                directory does not exist, it is created.
        """

        # Check if directory exists
        if not os.path.isdir(path + "dsprites"):
            print("Creating mnist Directory")
            os.mkdir(path + "dsprites")

        filename = "dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz"
        # Check if file exists
        if not os.path.exists(path + "dsprites/" + filename):
            td = time.time()
            url = "https://github.com/deepmind/dsprites-dataset/blob/master/" + filename
            urllib.request.urlretrieve(url, path + "dsprites/filename.npz")

    def load(path=None):
        """
        Parameters
        ----------
            path: str (optional)
                default ($DATASET_PATH), the path to look for the data and
                where the data will be downloaded if not present

        Returns
        -------

            images: array

            latent: array

            classes: array

        """

        if path is None:
            path = os.environ["DATASET_PATH"]

        download(path)

        t0 = time.time()
        dataset_zip = np.load("dsprites_ndarray_co1sh3sc6or40x32y32_64x64.npz")
        imgs = dataset_zip["imgs"]
        latents_values = dataset_zip["latents_values"]
        latents_classes = dataset_zip["latents_classes"]
        metadata = dataset_zip["metadata"][()]
        return imgs, latent_values, latents_classes



import os
import urllib.request
import time
import io
from scipy.io import loadmat
import mne
from tqdm import tqdm


class seizures_neonatal:
    """A dataset of neonatal EEG recordings with seizures annotations

    source: https://zenodo.org/record/2547147
   
    Neonatal seizures are a common emergency inthe neonatal intensive care unit
    (NICU). There are many questions yet to be answered regarding the
    temporal/spatial characteristics of seizures from different pathologies,
    response to medication, effects on neurodevelopment and optimal detection.
    This dataset contains EEG recordings from human neonates and the visual
    interpretation of the EEG by the human expert. Multi-channel EEG was
    recorded from 79 term neonates admitted to the neonatal intensive care unit
    (NICU) at the Helsinki University Hospital. The median recording duration
    was 74 minutes (IQR: 64 to 96 minutes). EEGs were annotated by three experts
    for the presence of seizures. An average of 460 seizures were annotated per
    expert in the dataset, 39 neonates had seizures by consensus and 22 were
    seizure free by consensus. The dataset can be used as a reference set of
    neonatal seizures, for the development of automated methods of seizure
    detection and other EEG analysis, as well as for the analysis of
    inter-observer agreement.
    """

    def download(path):

        if path is None:
            path = os.environ["DATASET_PATH"]

        # Load the dataset (download if necessary) and set
        # the class attributes.

        t = time.time()

        if not os.path.isdir(path + "seizures_neonatal"):

            print("\tCreating Directory")
            os.mkdir(path + "seizures_neonatal")

        if not os.path.exists(path + "seizures_neonatal/annotations_2017.mat"):
            print("Downloading Annotations")
            url = "https://zenodo.org/record/2547147/files/annotations_2017.mat?download=1"
            urllib.request.urlretrieve(
                url, path + "seizures_neonatal/annotations_2017.mat"
            )

        for i in tqdm(range(1, 80), ascii=True, desc="Downloading"):
            dest = "seizures_neonatal/eeg{}.edf".format(i)
            url = "https://zenodo.org/record/2547147/files/eeg{}.edf?download=1".format(
                i
            )
            if not os.path.exists(path + dest):
                urllib.request.urlretrieve(url, path + dest)

    def load(path=None):
        """
        Parameters
        ----------

        path: str (optional)
            a string where to load the data and download if not present

        Returns
        -------

        annotations: list
            the list of multichannel binary vectors representing
            the presence or absence of seizure, 3 channels due to
            3 expert annotations

        waveforms: list
            list of (channels, TIME) multichannel EEGs
        """
        if path is None:
            path = os.environ["DATASET_PATH"]

        seizures_neonatal.download(path)
        t = time.time()

        # load wavs
        annotations = loadmat(path + "seizures_neonatal/annotations_2017.mat")
        annotations = annotations["annotat_new"][0]

        # init. the data array
        waveforms = []
        for i in tqdm(range(1, 80), ascii=True):
            filename = path + "seizures_neonatal/eeg{}.edf".format(i)
            data = mne.io.read_raw_edf(filename)
            waveforms.append(data.get_data())

        return annotations, waveforms



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import numpy as np
from .. import tensor as T

__author__ = "Randall Balestriero"

_HERMITE = np.array(
    [[1, 0, -3, 2], [0, 0, 3, -2], [0, 1, -2, 1], [0, 0, -1, 1]], dtype="float32"
)


def upsample_1d(
    tensor, repeat, axis=-1, mode="constant", value=0.0, boundary_condition="periodic"
):
    """1-d upsampling of tensor

    allow to upsample a tensor by an arbitrary (integer) amount on a given
    axis by applying a univariate upsampling strategy.

    Parameters
    ----------

    tensor: tensor
        the input tensor to upsample

    repeat: int
        the amount of new values ot insert between each value

    axis: int
        the axis to upsample

    mode: str
        the type of upsample to perform (linear, constant, nearest)

    value: float (default=0)
        the value ot use for the case of constant upsampling

    """

    if axis == -1:
        axis = tensor.ndim - 1

    if repeat == 0:
        return tensor

    out_shape = list(tensor.shape)
    out_shape[axis] *= 1 + repeat

    if mode == "constant":
        zshape = list(tensor.shape)
        zshape.insert(axis + 1, repeat)
        tensor_aug = T.concatenate(
            [
                T.expand_dims(tensor, axis + 1),
                T.full(zshape, value, dtype=tensor.dtype),
            ],
            axis + 1,
        )

    elif mode == "nearest":
        if boundary_condition == "periodic":
            return T.roll(T.repeat(tensor, repeat + 1, axis), -repeat // 2, axis)
        else:
            raise NotImplemented

    elif mode == "linear":
        assert tensor.shape[axis] > 1
        zshape = [1] * (tensor.ndim + 1)
        zshape[axis + 1] = repeat
        coefficients = T.linspace(0, 1, repeat + 2)[1:-1].reshape(zshape)
        augmented_tensor = T.expand_dims(tensor, axis + 1)
        if boundary_condition == "periodic":
            interpolated = (
                augmented_tensor * (1 - coefficients)
                + T.roll(augmented_tensor, -1, axis) * coefficients
            )
        elif boundary_condition == "mirror":
            assert axis == tensor.ndim - 1
            other = T.index_update(
                T.roll(augmented_tensor, -1, axis),
                T.index[..., -1, :],
                augmented_tensor[..., -2, :],
            )
            interpolated = augmented_tensor * (1 - coefficients) + other * coefficients

        tensor_aug = T.concatenate([augmented_tensor, interpolated], axis + 1)

    return tensor_aug.reshape(out_shape)


def hermite_1d(samples, knots, values, derivatives):
    """Real interpolation with hermite cubic spline.

    Arguments
    ---------
        knots: array-like
            The knots onto the function is defined (derivative and
            antiderivative) tensor of knots can be given in which case
            the shape is (..., N_KNOTS) the first dimensions are treated
            independently.
        samples: array-like
            The points where the interpolation is required of shape
            (TIME). If the shape is more, the first dimensions must be
            broadcastable agains knots.
        values: array-like
            The real values of amplitude onto knots, same shape as knots.
        derivative: array-like
            The real values of derivatives onto knots, same shape as knots.

    Returns
    -------
        yi: array-like
            The interpolated real-valued function.
            :param derivatives:
    """

    # Concatenate coefficients onto shifted knots (..., N_KNOTS - 1, 2)
    adj_knots = T.stack([knots[..., :-1], knots[..., 1:]], axis=-1)
    adj_v = T.stack([values[..., :-1], values[..., 1:]], axis=-1)
    adj_d = T.stack([derivatives[..., :-1], derivatives[..., 1:]], axis=-1)

    # Define the full function y to interpolate (..., N_KNOTS - 1, 4)
    adj_vd = T.concatenate([adj_v, adj_d], axis=-1)

    # Extract Hermite polynomial coefficients (..., N_KNOTS - 1, 4)
    yh = T.matmul(adj_vd, _HERMITE)

    # Now we must apply a duplication over the number of knots, apply the
    # polynomial interpolation, and then mask for each region and sum
    # over the regions (..., N_KNOTS - 1, TIME)
    if samples.ndim == 1:
        samples_ = samples.reshape([1] * knots.ndim + [-1])
    else:
        samples_ = T.expand_dims(samples, -2)

    # we keep left and right and they will coincide for adjacent regions
    start = T.expand_dims(knots[..., :-1], -1)
    end = T.expand_dims(knots[..., 1:], -1)
    pos = (samples_ - start) / (end - start)
    mask = ((pos >= 0.0) * (pos <= 1.0)).astype("float32")
    mask = mask / T.maximum(1, mask.sum(-2, keepdims=True))

    # create the polynomial basis (..., N_KNOTS - 1, TIME, 4)
    polynome = T.expand_dims(pos, -1) ** T.arange(4)

    # apply mask
    mask_polynome = polynome * T.expand_dims(mask, -1)

    # linearly combine to produce interpolation
    return (T.expand_dims(yh, -2) * mask_polynome).sum(axis=(-3, -1))


def hermite_2d(values, n_x, n_y):
    """
    TODO: test and finalize this

    Parameters
    ----------

    values: array-like
        the values, and 2 directional derivatives and the cross derivative
        for the 4 knots per region, hence it should be of shape n,N,M,4
        values vx vy vxy

    n_x: int
        the number of points in x per region

    n_y: int
        the number of points in y per region

    Returns
    -------

    interpolation: array-like

    """
    n, N, M = values.shape[:3]
    R_N = N - 1
    R_M = M - 1
    patches = T.extract_image_patches(values, (2, 2, 1))  # (n, R_N, R_M, 2, 2, 4)

    F = T.concatenate(
        [
            T.concatenate([patches[..., 0], patches[..., 1]], -1),
            T.concatenate([patches[..., 2], patches[..., 3]], -1),
        ],
        axis=-2,
    )  # (n, R_N, R_M, 4, 4)

    M = T.Variable(
        array([[1, 0, 0, 0], [0, 0, 1, 0], [-3, 3, -2, -1], [2, -2, 1, 1]]).astype(
            "float32"
        ),
        trainable=False,
    )

    MFM = T.einsum("xnmij,ai,bj->xnmab", F, M, M)  # (n, R_N, R_M, 4, 4)

    # filter shape is (n_x,n_y,(n-1)*self.R_N,(n-1)*self.R_M)
    t_x = T.linspace(float32(0), float32(1), int32(n_x - 1))
    t_y = T.linspace(float32(0), float32(1), int32(n_y - 1))
    x = T.pow(t_x, T.arange(4)[:, None])  # (4,T-1)
    y = T.pow(t_y, T.arange(4)[:, None])  # (4,T-1)
    values = T.einsum("xnmij,ia,jb->xnamb", MFM, x, y)
    return T.reshape(values, (n, (R_N) * (n - 1), (R_M) * (n - 1)))
